function [x,r,normR,residHist, errHist] = FBP( A, b,errFcn, opts )
% x = NEW( A, b )
%   uses the Compressive Sampling Matched Pursuit (New) algorithm
%   (see Needell and Tropp's 2008 paper http://arxiv.org/abs/0803.2392 )
%   to estimate the solution to the equation
%       b = A*x   
%   where there is prior information that x is sparse.
%
%   "A" may be a matrix, or it may be a cell array {Af,At}
%   where Af and At are function handles that compute the forward and transpose
%   multiplies, respectively. If the function handles are provided,
%   the the least-squares step is performed using LSQR (use could also use
%   CG on the normal equations, or other special variants).
%
% [x,r,normR,residHist,errHist] = New( A, b errFcn, opts )
%   is the full version.
% Outputs:
%   'x' is the k-sparse estimate of the unknown signal
%   'r' is the residual b - A*x
%   'normR' = norm(r)
%   'residHist'     is a vector with normR from every iteration
%   'errHist'       is a vector with the outout of errFcn from every iteration
%
% Inputs:
%   'A'     is the measurement matrix
%   'b'     is the vector of observations
%                   solve the internal least-squares problem uniquely.
%   'errFcn'    (optional; set to [] to ignore) is a function handle
%              which will be used to calculate the error; the output
%              should be a scalar
%   'opts'  is a structure with more options, including:
%       .printEvery = is an integer which controls how often output is printed
%       .maxiter    = maximum number of iterations
%       .normTol    = desired size of norm(residual). This is also
%                       used to detect convergence when the residual
%                       has stopped decreasing in norm
%       .LSQR_tol   = when "A" is a set of function handles, this controls
%                       the tolerance in the iterative solver. For compatibility
%                       with older versions, the fieldname "cg_tol" is also OK.
%       .LSQR_maxit = maximum number of steps in the iterative solver. For compatibility
%                       with older versions, the fieldname "cg_maxit" is also OK.
%                       N.B. "CG" stands for conjugate gradient, but this code
%                            actually uses the LSQR solver.
%       .HSS        = if true, use the variant of NEW that is similar to
%                       HHS Pursuit (see appendix A.2 of Needell/Tropp paper). Recommended.
%       .two_solves = if true, uses the variant of NEW that re-solves
%                       on the support of size '(M/2)-1' at every iteration
%                       (see appendix). This can be used with or without HSS variant.
%       .addm       = the number of new entries to add each time. By default
%                       this is M/2 (this was what is used in the paper).
%       .support_tol = when adding the (addm) atoms wwhich minimize the
%                       least square method
%
%
size_m=round((size(A,1)*0.3));
% select_ind =round((size(A,1)/3));
 select_ind =size_m-1;

if nargin < 5, opts = []; end
if ~isempty(opts) && ~isstruct(opts)
    error('"opts" must be a structure');
end

function out = setOpts( field, default )
    if ~isfield( opts, field )
        opts.(field)    = default;
    end
    out = opts.(field);
end

printEvery  = setOpts( 'printEvery', 1000 );
maxiter     = setOpts( 'maxiter', size(A,1) );
normTol     = setOpts( 'normTol', 1e-10 );
cg_tol      = setOpts( 'cg_tol', 1e-8 );
cg_maxit    = setOpts( 'cg_maxit', 20 );
% Allow some synonyms
cg_tol      = setOpts( 'LSQR_tol', cg_tol );
cg_maxit    = setOpts( 'LSQR_maxit', cg_maxit );
HSS         = setOpts( 'HSS', true );
TWO_SOLVES  = setOpts( 'two_solves', true );
%**********************************************************
addm        = round(setOpts( 'addm', size_m));
%***********************************************************
support_tol = setOpts( 'support_tol', 1e-10 );
if iscell(A)
    LARGESCALE  = true;
    Af  = A{1};
    At  = A{2};     % we don't really need this...
else
    LARGESCALE  = false;
    Af  = @(x) A*x;
    %**********************************************************
    At  = @(x) pinv(A)*x;      %Basic function
    %**********************************************************
end
% -- Intitialize --
% start at x = 0, so r = b - A*x = b
r           = b;
Ar          = At(r);
N           = size(Ar,1);       % number of atoms
M           = size(r,1);        % size of atoms
x           = zeros(N,1);
ind_m       = [];
%**********************************************************
residHist   = zeros(size_m,1);
errHist     = zeros(size_m,1);
%**********************************************************
fprintf('Iter,   |T|,  Resid');
if ~isempty(errFcn)
    fprintf(',   Error');
end
if LARGESCALE
    fprintf(',   LSQR iterations and norm(residual)');
end
fprintf('\n');

if LARGESCALE  
    LSQR_ALG    = @(RHS,Afcn,x0) lsqr(Afcn,RHS,cg_tol,cg_maxit,[],[],x0 );  
end
for kk = 1:maxiter 
    % -- Step 1: find new index and atom to add
    %**********************************************************
    y_sort      = sort( abs(Ar),'descend');
    cutoff      = y_sort(addm); % addm is typically m/2
    cutoff      = max( cutoff, support_tol );
    ind_new     = find( abs(Ar) >= cutoff );
    %**********************************************************
    % -- Merge:
    T    = union( ind_new, ind_m );
    % -- Step 2: update residual
    if HSS
        RHS     = r; % where r = b - A*x, so we'll need to add in "x" later
        x_warmstart     = zeros(length(T),1);
    else
        RHS     = b;
        x_warmstart     = x(T);
    end
    % -- solve for x on the suppor set "T"
    if LARGESCALE
        % Could do CG on the normal equations...
        % Or, use LSQR:
%         x_T = LSQR_ALG(Afcn,RHS,cg_tol,cg_maxit);
        % use an initial guess to warm-start the solver
        Afcn = @(x,mode) partialA( N, T, Af, At, x, mode );
        [x_T,flag,relres,CGiter] = LSQR_ALG(RHS,Afcn,x_warmstart);  
        
    else
        x_T = A(:,T)\RHS;   % more efficient; equivalent to pinv when underdetermined.
%         x_T     = pinv( A(:,T) )*RHS;
    end
    if HSS
        % HSS variation of New
        x_new       = zeros(N,1);
        x_new(T)    = x_T;
        x           = x + x_new;    % this is the key extra step in HSS
        %**********************************************************
        cutoff      = findCutoff( x, select_ind);
        %**********************************************************
        x           = x.*( abs(x) >= cutoff );
        ind_m       = find(x);
        
        if TWO_SOLVES
            if LARGESCALE
                Afcn = @(x,mode) partialA( N, ind_m, Af, At, x, mode );
                [x_T2,flag,relres,CGiter] = LSQR_ALG(b,Afcn,x(ind_m)); % not using "r", just using "b"
            else
                x_T2     = A(:,ind_m)\b;
            end
            x( ind_m )  = x_T2;
        end
        
        % update r
        r_old   = r;
        r       = b - Af(x);
    else
        % Standard New
        % Note: this is implemented *slightly* more efficiently
        %   that the HSS variation
    
        % Prune x to keep only "(m/2)-1" entries:
        %**********************************************************
        cutoff  = findCutoff(x_T, select_ind);
        %**********************************************************
        Tk      = find( abs(x_T) >= cutoff );
        % This is assuming there are no ties. If there are,
        %    from a practical standpoint, it probably doesn't
        %    matter much what you do. So out of laziness, we don't worry about it.
        ind_m   = T(Tk);
        x       = 0*x;
        x( ind_m ) = x_T( Tk );
        
        
        if TWO_SOLVES
            if LARGESCALE
                Afcn = @(x,mode) partialA( N, ind_m, Af, At, x, mode );
                [x_T2,flag,relres,CGiter] = LSQR_ALG(b,Afcn,x(ind_m));
            else
                x_T2     = A(:,ind_m)\b;
            end
            x( ind_m )  = x_T2;
        end
        
        % Update x and r
        r_old   = r;
        if LARGESCALE
            r   = b - Af( x );
        else
            % don't do a full matrix-vector multiply, just use necessary columns
            r   = b - A(:,ind_m)*x_T( Tk );
        end
    end

    
    % -- Print some info --
    PRINT   = ( ~mod( kk, printEvery ) || kk == maxiter );
    normR   = norm(r);
    STOP    = false;
    if normR < normTol || norm( r - r_old ) < normTol
        STOP    = true;
        PRINT   = true;
    end
    
    
    if ~isempty(errFcn)
        er  = errFcn(x);
        errHist(kk)     = er;
    end
    if PRINT
        fprintf('%4d, %4d, %.2e', kk, length(T), normR);
        if ~isempty(errFcn)
            fprintf(', %.2e',er);
        end
        if LARGESCALE
            fprintf(',  %d,  %.2e', CGiter,relres);
        end
        fprintf('\n');
    end
        
    residHist(kk)   = normR;
    
    if STOP
        disp('Reached stopping criteria');
        break;
    end

    if kk < maxiter
        Ar  = At(r); % prepare for next round
    end
    
end

end % -- end of main function

function tau = findCutoff( x, k )
% finds the appropriate cutoff such that after hard-thresholding,
% "x" will be k-sparse
x   = sort( abs(x),'descend');
if k > length(x)
    tau = x(end)*.999;
else
    tau  = x(k);
end
end


